# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import importlib
import re
from collections.abc import Callable
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Union

from pydantic import BaseModel

from .config import get_config
from .schema_generation import (
    create_abstract_eval_schema,
    create_apply_schema,
    create_autodiff_schema,
)
from .tree_transforms import get_at_path


def load_module_from_path(path: Union[Path, str]) -> ModuleType:
    """Load a module from a file path."""
    path = Path(path)

    if not path.is_file():
        raise ImportError(f"Could not load module from {path} (is not a file)")

    module_name = path.stem
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None:
        raise ImportError(f"Could not load module from {path}")

    module = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as exc:
        raise ImportError(f"Could not load module from {path}") from exc
    return module


def get_supported_endpoints(api_module: ModuleType) -> tuple[str, ...]:
    """Get available Tesseract functions.

    Returns:
        All optional function names defined by the Tesseract.
    """
    optional_funcs = {
        "jacobian",
        "jacobian_vector_product",
        "vector_jacobian_product",
        "abstract_eval",
    }
    return tuple(func for func in optional_funcs if hasattr(api_module, func))


def get_tesseract_api() -> ModuleType:
    """Import tesseract_api.py file."""
    return load_module_from_path(get_config().tesseract_api_path)


def _validate_ad_input(
    inputs: BaseModel,
    input_keys: set[str],
    output_keys: set[str],
    endpoint_name: str,
    tangent_cotangent_vector: Optional[dict[str, Any]] = None,
):
    """Raise an exception if the input of an autodiff function does not conform to given input keys."""
    # Could be moved to a model validator
    for input_key in input_keys:
        if not re.match(r"^[a-zA-Z0-9_.\[\]{}]+$", input_key):
            raise RuntimeError(
                f"Error when validating input of function {endpoint_name}:\n"
                f"Invalid input key {input_key}."
            )

        try:
            get_at_path(inputs, input_key)
        except Exception as exc:
            raise RuntimeError(
                f"Error when validating input of function {endpoint_name}:\n"
                f"Could not find input path {input_key} in inputs."
            ) from exc

    if endpoint_name == "jacobian_vector_product":
        # Tangent vector needs same keys as input_keys
        if set(tangent_cotangent_vector.keys()) != input_keys:
            raise RuntimeError(
                f"Error when validating input of function {endpoint_name}:\n"
                f"Expected tangent vector with keys {input_keys}, got {set(tangent_cotangent_vector.keys())}."
            )

    if endpoint_name == "vector_jacobian_product":
        # Cotangent vector needs same keys as output_keys
        if set(tangent_cotangent_vector.keys()) != output_keys:
            raise RuntimeError(
                f"Error when validating input of function {endpoint_name}:\n"
                f"Expected cotangent vector with keys {output_keys}, got {set(tangent_cotangent_vector.keys())}."
            )


def _validate_ad_output(
    input: Any,
    output: Any,
    input_keys: set[str],
    output_keys: set[str],
    endpoint_name: str,
):
    """Raise an exception if the output structure of an autodiff function does not conform to given keys."""
    if not isinstance(output, dict):
        raise RuntimeError(
            f"Error when validating output of function {endpoint_name}:\n"
            f"Expected output to be a dictionary, got {type(output)}"
        )

    if endpoint_name == "jacobian":
        if output_keys != output.keys():
            raise RuntimeError(
                "Error when validating output of jacobian:\n"
                f"Expected keys {output_keys} in output; got {set(output.keys())}"
            )
        for subkey, subout in output.items():
            if not isinstance(subout, dict):
                raise RuntimeError(
                    "Error when validating output of jacobian:\n"
                    f"Expected output with structure {{{tuple(output_keys)}: {{{tuple(input_keys)}: ...}}}}, "
                    f"got unexpected type {type(subout)} for output key {subkey}."
                )
            if input_keys != subout.keys():
                raise RuntimeError(
                    "Error when validating output of jacobian:\n"
                    f"Expected output with structure {{{tuple(output_keys)}: {{{tuple(input_keys)}: ...}}}}, "
                    f"got {set(subout.keys())} for output key {subkey}."
                )

    elif endpoint_name == "jacobian_vector_product":
        if output_keys != output.keys():
            raise RuntimeError(
                "Error when validating output of jacobian_vector_product:\n"
                f"Expected keys {output_keys} in output; got {set(output.keys())}"
            )

    elif endpoint_name == "vector_jacobian_product":
        if input_keys != output.keys():
            raise RuntimeError(
                "Error when validating output of vector_jacobian_product:\n"
                f"Expected keys {input_keys} in output; got {set(output.keys())}"
            )
    else:
        raise RuntimeError(f"Unknown endpoint name {endpoint_name}")


def create_endpoints(api_module: ModuleType) -> list[Callable]:
    """Create the Tesseract API endpoints.

    This ensures proper type annotations, signatures, and validation for the external-facing API.

    Args:
        api_module: The Tesseract API module.

    Returns:
        A tuple of all Tesseract API endpoints as callables.
    """
    supported_functions = get_supported_endpoints(api_module)

    endpoints = []

    def assemble_docstring(wrapped_func: Callable):
        """Decorator to assemble a docstring from multiple functions."""

        def inner(otherfunc: Callable):
            doc_parts = []
            if otherfunc.__doc__:
                doc_parts.append(otherfunc.__doc__)

            if wrapped_func.__doc__:
                doc_parts.append(wrapped_func.__doc__)

            otherfunc.__doc__ = "\n\n".join(doc_parts)
            return otherfunc

        return inner

    ApplyInputSchema, ApplyOutputSchema = create_apply_schema(
        api_module.InputSchema, api_module.OutputSchema
    )

    @assemble_docstring(api_module.apply)
    def apply(payload: ApplyInputSchema) -> ApplyOutputSchema:
        """Apply the Tesseract to the input data."""
        out = api_module.apply(payload.inputs)
        if isinstance(out, api_module.OutputSchema):
            out = out.model_dump()
        return ApplyOutputSchema.model_validate(out)

    endpoints.append(apply)

    if "jacobian" in supported_functions:
        JacobianInputSchema, JacobianOutputSchema = create_autodiff_schema(
            api_module.InputSchema, api_module.OutputSchema, ad_flavor="jacobian"
        )

        @assemble_docstring(api_module.jacobian)
        def jacobian(payload: JacobianInputSchema) -> JacobianOutputSchema:
            """Computes the Jacobian of the Tesseract.

            Differentiates ``jac_outputs`` with respect to ``jac_inputs``, at the point ``inputs``.
            """
            _validate_ad_input(
                payload.inputs, payload.jac_inputs, payload.jac_outputs, "jacobian"
            )
            out = api_module.jacobian(
                payload.inputs, payload.jac_inputs, payload.jac_outputs
            )
            _validate_ad_output(
                payload.inputs, out, payload.jac_inputs, payload.jac_outputs, "jacobian"
            )
            return JacobianOutputSchema.model_validate(out)

        endpoints.append(jacobian)

    if "jacobian_vector_product" in supported_functions:
        JVPInputSchema, JVPOutputSchema = create_autodiff_schema(
            api_module.InputSchema, api_module.OutputSchema, ad_flavor="jvp"
        )

        @assemble_docstring(api_module.jacobian_vector_product)
        def jacobian_vector_product(payload: JVPInputSchema) -> JVPOutputSchema:
            """Compute the Jacobian vector product of the Tesseract at the input data.

            Evaluates the Jacobian vector product between the Jacobian given by ``jvp_outputs``
            with respect to ``jvp_inputs`` at the point ``inputs`` and the given tangent vector.
            """
            _validate_ad_input(
                payload.inputs,
                payload.jvp_inputs,
                payload.jvp_outputs,
                "jacobian_vector_product",
                tangent_cotangent_vector=payload.tangent_vector,
            )
            out = api_module.jacobian_vector_product(
                payload.inputs,
                payload.jvp_inputs,
                payload.jvp_outputs,
                payload.tangent_vector,
            )
            _validate_ad_output(
                payload.inputs,
                out,
                payload.jvp_inputs,
                payload.jvp_outputs,
                "jacobian_vector_product",
            )
            return JVPOutputSchema.model_validate(out)

        endpoints.append(jacobian_vector_product)

    if "vector_jacobian_product" in supported_functions:
        VJPInputSchema, VJPOutputSchema = create_autodiff_schema(
            api_module.InputSchema, api_module.OutputSchema, ad_flavor="vjp"
        )

        @assemble_docstring(api_module.vector_jacobian_product)
        def vector_jacobian_product(payload: VJPInputSchema) -> VJPOutputSchema:
            """Compute the Jacobian vector product of the Tesseract at the input data.

            Computes the vector Jacobian product between the Jacobian given by ``vjp_outputs``
            with respect to ``vjp_inputs`` at the point ``inputs`` and the given cotangent vector.
            """
            _validate_ad_input(
                payload.inputs,
                payload.vjp_inputs,
                payload.vjp_outputs,
                "vector_jacobian_product",
                tangent_cotangent_vector=payload.cotangent_vector,
            )
            out = api_module.vector_jacobian_product(
                payload.inputs,
                payload.vjp_inputs,
                payload.vjp_outputs,
                payload.cotangent_vector,
            )
            _validate_ad_output(
                payload.inputs,
                out,
                payload.vjp_inputs,
                payload.vjp_outputs,
                "vector_jacobian_product",
            )
            return VJPOutputSchema.model_validate(out)

        endpoints.append(vector_jacobian_product)

    def health() -> dict[str, Any]:
        """Get health status of the Tesseract instance."""
        return {"status": "ok"}

    endpoints.append(health)

    def input_schema() -> dict[str, Any]:
        """Get input schema for tesseract apply function."""
        return api_module.InputSchema.model_json_schema()

    endpoints.append(input_schema)

    def output_schema() -> dict[str, Any]:
        """Get output schema for tesseract apply function."""
        return api_module.OutputSchema.model_json_schema()

    endpoints.append(output_schema)

    if "abstract_eval" in supported_functions:
        AbstractEvalInputSchema, AbstractEvalOutputSchema = create_abstract_eval_schema(
            api_module.InputSchema, api_module.OutputSchema
        )

        @assemble_docstring(api_module.abstract_eval)
        def abstract_eval(payload: AbstractEvalInputSchema) -> AbstractEvalOutputSchema:
            """Perform abstract evaluation of the Tesseract on the input data."""
            out = api_module.abstract_eval(payload.inputs)
            return AbstractEvalOutputSchema.model_validate(out)

        endpoints.append(abstract_eval)

    return endpoints



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""This module provides a command-line interface for interacting with the Tesseract runtime."""

import contextlib
import io
import os
import sys
from collections.abc import Callable, Generator
from pathlib import Path
from textwrap import dedent
from typing import Any, Optional

import click
import typer

from tesseract_core.runtime.config import get_config
from tesseract_core.runtime.core import create_endpoints, get_tesseract_api
from tesseract_core.runtime.file_interactions import (
    SUPPORTED_FORMATS,
    guess_format_from_path,
    load_bytes,
    output_to_bytes,
    read_from_path,
    write_to_path,
)
from tesseract_core.runtime.serve import create_rest_api
from tesseract_core.runtime.serve import serve as serve_


class SpellcheckedTyperGroup(typer.core.TyperGroup):
    """A Typer group that suggests similar commands if a command is not found."""

    def get_command(self, ctx: click.Context, invoked_command: str) -> Any:
        """Get a command from the Typer group, suggesting similar commands if the command is not found."""
        import difflib

        possible_commands = self.list_commands(ctx)
        if invoked_command not in possible_commands:
            close_match = difflib.get_close_matches(
                invoked_command, possible_commands, n=1, cutoff=0.6
            )
            if close_match:
                raise click.UsageError(
                    f"No such command '{invoked_command}'. Did you mean '{close_match[0]}'?",
                    ctx,
                )
        return super().get_command(ctx, invoked_command)


app = typer.Typer(name="tesseract-runtime", cls=SpellcheckedTyperGroup)


def _prettify_docstring(docstr: str) -> str:
    """Enforce consistent indentation level of docstrings."""
    # First line is not indented, the rest is -> leads to formatting issues
    docstring_lines = docstr.split("\n")
    dedented_lines = dedent("\n".join(docstring_lines[1:]))
    return "\n".join([docstring_lines[0].lstrip(), dedented_lines])


def _parse_arg_callback(
    ctx: Any, param: Any, value: Any
) -> tuple[dict[str, Any], Optional[Path]]:
    """Click callback to parse Tesseract input arguments provided in the CLI.

    Returns a tuple of the parsed value and the base directory of the input path, if any.
    """
    base_dir = None

    if not isinstance(value, str):
        # Passthrough, probably a default value
        return value, base_dir

    if value.startswith("@"):
        base_dir = Path(value[1:]).parent
        value_format = guess_format_from_path(value[1:])
        try:
            value_bytes = read_from_path(value[1:])
        except Exception as e:
            raise click.BadParameter(f"Could not read data from path {value}") from e
    else:
        # Data given directly via the CLI is always in JSON format
        value_format = "json"
        value_bytes = value.encode()

    try:
        decoded_value = load_bytes(value_bytes, value_format)
    except Exception as e:
        raise click.BadParameter(
            f"Could not decode data using format {value_format}"
        ) from e

    return decoded_value, base_dir


@app.callback(
    no_args_is_help=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)
def main_callback() -> None:
    """Invoke the Tesseract runtime.

    The Tesseract runtime can be configured via environment variables; for example,
    ``export TESSERACT_RUNTIME_PORT=8080`` sets the port to use for ``tesseract serve`` to 8080.
    """
    pass


tesseract_runtime = typer.main.get_command(app)


def _schema_to_docstring(schema: Any, current_indent: int = 0) -> str:
    """Convert a Pydantic schema to a human-readable docstring."""
    docstring = []
    indent = " " * current_indent

    if not hasattr(schema, "model_fields"):
        return ""

    is_root = schema.model_fields.keys() == {"root"}

    for field, field_info in schema.model_fields.items():
        if is_root:
            docline = f"{indent}{field_info.description}"
        else:
            docline = f"{indent}{field}: {field_info.description}"

        if (
            field_info.default is not None
            and str(field_info.default) != "PydanticUndefined"
        ):
            docline = f"{docline} [default: {field_info.default}]"

        docstring.append(docline)

        if hasattr(field_info.annotation, "model_fields"):
            docstring.append(
                _schema_to_docstring(field_info.annotation, current_indent + 4)
            )

    return "\n".join(docstring)


@tesseract_runtime.command()
def check() -> None:
    """Check whether the Tesseract API can be imported."""
    # raises an exception if the API cannot be imported
    get_tesseract_api()
    typer.echo("✅ Tesseract API check successful ✅")


@tesseract_runtime.command()
@click.option("-p", "--port", default=8000, help="Port number")
@click.option("-h", "--host", default="0.0.0.0", help="Host IP address")
@click.option("-w", "--num-workers", default=1, help="Number of worker processes")
def serve(host: str, port: int, num_workers: int) -> None:
    """Start running this Tesseract's web server."""
    serve_(host=host, port=port, num_workers=num_workers)


def _create_user_defined_cli_command(
    user_function: Callable, out_stream: Optional[io.IOBase]
) -> click.Command:
    """Creates a click command which sends requests to Tesseract endpoints.

    We need to do this dynamically, as we want to generate docs and usage
    from the Tesseract api's signature and docstrings.

    Args:
        user_function: The user-defined function to create a CLI command for.
        out_stream: The default output stream to write to. If None, defaults to
            sys.stdout at the time of invocation.
    """
    InputSchema = user_function.__annotations__.get("payload", None)
    OutputSchema = user_function.__annotations__.get("return", None)

    options = []

    if InputSchema is not None:
        options.append(
            click.Argument(
                ["payload"],
                type=click.STRING,
                required=True,
                metavar="JSON_PAYLOAD",
                callback=_parse_arg_callback,
            )
        )

    options.extend(
        [
            click.Option(
                ["-o", "--output-path"],
                type=click.STRING,
                help=(
                    "Output path to write the result to, such as local directory or S3 URI "
                    "(may be anything supported by fsspec). [default: write to stdout]"
                ),
                default=None,
            ),
            click.Option(
                ["-f", "--output-format"],
                type=click.Choice(SUPPORTED_FORMATS),
                help="Output format to write results in.",
                default="json",
            ),
        ]
    )

    def _callback_wrapper(
        output_path: Optional[str],
        output_format: SUPPORTED_FORMATS,
        **optional_args: Any,
    ):
        if output_format == "json+binref" and output_path is None:
            raise ValueError("--output-path must be specified for json+binref format")

        out_stream_ = out_stream or sys.stdout

        user_function_args = {}

        if InputSchema is not None:
            payload, base_dir = optional_args["payload"]
            user_function_args["payload"] = InputSchema.model_validate(
                payload, context={"base_dir": base_dir}
            )

        result = user_function(**user_function_args)
        result = output_to_bytes(result, output_format, output_path)

        if output_path is None:
            # write raw bytes to out_stream.buffer to support binary data (which may e.g. be piped)
            out_stream_.buffer.write(result)
            out_stream_.flush()
        else:
            format = output_format.split("+", maxsplit=1)[0]
            write_to_path(result, f"{output_path}/results.{format}")

    function_name = user_function.__name__.replace("_", "-")

    # Assemble docstring
    function_doc = [_prettify_docstring(user_function.__doc__)]

    if InputSchema is not None and hasattr(InputSchema, "model_fields"):
        function_doc.append(
            "\nFirst argument is the payload, which should be a JSON object with the following structure."
        )

        # \b\n disables click's automatic formatting
        function_doc.append("\n\n\b\nInput schema:")
        function_doc.append(_schema_to_docstring(InputSchema, current_indent=4))

    if OutputSchema is not None and hasattr(OutputSchema, "model_fields"):
        function_doc.append("\n\n\b\nReturns:")
        function_doc.append(_schema_to_docstring(OutputSchema, current_indent=4))

    command = click.Command(
        function_name,
        short_help=f"Call the Tesseract function {function_name}.",
        help="\n".join(function_doc),
        callback=_callback_wrapper,
        params=options,
    )

    return command


def _add_user_commands_to_cli(
    group: click.Group, out_stream: Optional[io.IOBase]
) -> click.Group:
    tesseract_package = get_tesseract_api()
    endpoints = create_endpoints(tesseract_package)

    openapi_schema_ = create_rest_api(tesseract_package).openapi()

    def openapi_schema() -> dict:
        """Get the openapi.json schema."""
        return openapi_schema_

    endpoints.append(openapi_schema)

    for func in endpoints:
        group.add_command(_create_user_defined_cli_command(func, out_stream))

    return group


@contextlib.contextmanager
def stdout_to_stderr() -> Generator:
    """Redirect stdout to stderr at OS level."""
    orig_stdout = os.dup(sys.stdout.fileno())
    sys.stdout.flush()
    os.dup2(sys.stderr.fileno(), sys.stdout.fileno())
    try:
        yield os.fdopen(orig_stdout, "w", closefd=False)
    finally:
        sys.stdout.flush()
        os.dup2(orig_stdout, sys.stdout.fileno())


def main() -> None:
    """Entrypoint for the command line interface."""
    # Redirect stdout to stderr to avoid mixing any output with the JSON response.
    with stdout_to_stderr() as orig_stdout:
        # Fail as fast as possible if the Tesseract API path is not set
        api_path = get_config().tesseract_api_path
        if not api_path.is_file():
            print(
                f"Tesseract API file '{api_path}' does not exist. "
                "Please ensure it is a valid file, or set the TESSERACT_API_PATH "
                "environment variable to the path of your Tesseract API file.\n"
                "\n"
                "Example:\n"
                "    $ export TESSERACT_API_PATH=/path/to/your/tesseract_api.py\n"
                "\n"
                "Aborted.",
                file=sys.stderr,
            )
            sys.exit(1)
        cli = _add_user_commands_to_cli(tesseract_runtime, out_stream=orig_stdout)
        cli(auto_envvar_prefix="TESSERACT_RUNTIME")


if __name__ == "__main__":
    main()



import numpy as np
import pytest
from pydantic import ValidationError

from tesseract_core import Tesseract
from tesseract_core.sdk.tesseract import (
    HTTPClient,
    _decode_array,
    _encode_array,
    _tree_map,
)


@pytest.fixture
def mock_serving(mocker):
    serve_mock = mocker.patch("tesseract_core.sdk.engine.serve")
    serve_mock.return_value = "proj-id-123"

    subprocess_run_mock = mocker.patch("subprocess.run")
    subprocess_run_mock.return_value.stdout = (
        '{"ID": "abc1234", "Publishers":[{"PublishedPort": 54321}]}'
    )

    teardown_mock = mocker.patch("tesseract_core.sdk.engine.teardown")
    return {
        "serve_mock": serve_mock,
        "subprocess_run_mock": subprocess_run_mock,
        "teardown_mock": teardown_mock,
    }


@pytest.fixture
def mock_clients(mocker):
    mocker.patch("tesseract_core.sdk.tesseract.HTTPClient._run_tesseract")


def test_Tesseract_init():
    # Instantiate with a url
    t = Tesseract(url="localhost")

    # The attributes for local Tesseracts should not be set
    assert not hasattr(t, "image")
    assert not hasattr(t, "gpus")
    assert not hasattr(t, "volumes")


def test_Tesseract_from_image():
    # Object is built and has the correct attributes set
    t = Tesseract.from_image("sometesseract:0.2.3", volumes=["/my/files"], gpus=["all"])

    assert t.image == "sometesseract:0.2.3"
    assert t.gpus == ["all"]
    assert t.volumes == ["/my/files"]

    # Let's also check that stuff we don't expect there is not there
    assert not hasattr(t, "url")


def test_Tesseract_schema_methods(mocker, mock_serving):
    mocked_run = mocker.patch("tesseract_core.sdk.tesseract.HTTPClient._run_tesseract")
    mocked_run.return_value = {"#defs": {"some": "stuff"}}

    with Tesseract.from_image("sometesseract:0.2.3") as t:
        input_schema = t.input_schema
        output_schema = t.output_schema
        openapi_schema = t.openapi_schema

    assert input_schema == output_schema == openapi_schema == mocked_run.return_value


def test_serve_lifecycle(mock_serving, mock_clients):
    t = Tesseract.from_image("sometesseract:0.2.3")

    with t:
        pass

    mock_serving["serve_mock"].assert_called_with(
        ["sometesseract:0.2.3"], port="", volumes=None, gpus=None
    )

    mock_serving["teardown_mock"].assert_called_with("proj-id-123")

    # check that the same Tesseract obj cannot be used to instantiate two containers
    with pytest.raises(RuntimeError):
        with t:
            with t:
                pass


def test_HTTPClient_run_tesseract(mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"result": [4, 4, 4]}
    mock_response.raise_for_status = mocker.Mock()

    mocked_request = mocker.patch(
        "requests.request",
        return_value=mock_response,
    )

    client = HTTPClient("somehost")

    out = client._run_tesseract("apply", {"inputs": {"a": 1}})

    assert out == {"result": [4, 4, 4]}
    mocked_request.assert_called_with(
        method="POST",
        url="http://somehost/apply",
        json={"inputs": {"a": 1}},
    )


def test_HTTPClient__run_tesseract_raises_validation_error(mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {
        "detail": [
            {
                "type": "missing",
                "loc": ["body", "inputs", "a"],
                "msg": "Field required",
                "input": {"whoops": "whatever"},
            },
            {
                "type": "missing",
                "loc": ["body", "inputs", "b"],
                "msg": "Field required",
                "input": {"whoops": "whatever"},
            },
            {
                "type": "extra_forbidden",
                "loc": ["body", "inputs", "whoops"],
                "msg": "Extra inputs are not permitted",
                "input": "whatever",
            },
        ]
    }
    mock_response.status_code = 422

    mocker.patch(
        "requests.request",
        return_value=mock_response,
    )

    client = HTTPClient("somehost")

    with pytest.raises(ValidationError):
        client._run_tesseract("apply", {"inputs": {"whoops": "whatever"}})


@pytest.mark.parametrize(
    "b64, expected_data",
    [
        (True, {"buffer": "AACAPwAAAEAAAEBA", "encoding": "base64"}),
        (False, {"buffer": [1.0, 2.0, 3.0], "encoding": "raw"}),
    ],
)
def test_encode_array(b64, expected_data):
    a = np.array([1.0, 2.0, 3.0], dtype="float32")

    encoded = _encode_array(a, b64=b64)

    assert encoded["shape"] == (3,)
    assert encoded["dtype"] == "float32"
    assert encoded["data"] == expected_data


@pytest.mark.parametrize(
    "encoded, expected",
    [
        (
            {
                "shape": (3,),
                "dtype": "float32",
                "data": {"buffer": [1.0, 2.0, 3.0], "encoding": "raw"},
            },
            [1.0, 2.0, 3.0],
        ),
        (
            {
                "shape": (3,),
                "dtype": "float32",
                "data": {"buffer": "AACAPwAAAEAAAEBA", "encoding": "base64"},
            },
            [1.0, 2.0, 3.0],
        ),
    ],
)
def test_decode_array(encoded, expected):
    decoded = _decode_array(encoded)
    assert np.all(decoded == expected)


def test_tree_map():
    tree = {
        "a": [10, 20],
        "b": {"c": np.array([30])},
        "d": {"e": np.array([1.0, 2.0, 3.0])},
        "f": "hello",
    }

    encoded = _tree_map(_encode_array, tree, is_leaf=lambda x: hasattr(x, "shape"))

    assert encoded == {
        "a": [10, 20],
        "b": {
            "c": {
                "shape": (1,),
                "dtype": "int64",
                "data": {"buffer": "HgAAAAAAAAA=", "encoding": "base64"},
            }
        },
        "d": {
            "e": {
                "shape": (3,),
                "dtype": "float64",
                "data": {
                    "buffer": "AAAAAAAA8D8AAAAAAAAAQAAAAAAAAAhA",
                    "encoding": "base64",
                },
            }
        },
        "f": "hello",
    }



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import ast
from pathlib import Path
from typing import Annotated, NamedTuple

import yaml
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    Strict,
)
from pydantic import ValidationError as PydanticValidationError


class _ApiObject(NamedTuple):
    name: str
    expected_type: type
    num_args: int | None = None
    optional: bool = False


EXPECTED_OBJECTS = (
    _ApiObject("apply", ast.FunctionDef, 1),
    _ApiObject("InputSchema", ast.ClassDef),
    _ApiObject("OutputSchema", ast.ClassDef),
    _ApiObject("jacobian", ast.FunctionDef, 3, optional=True),
    _ApiObject("jacobian_vector_product", ast.FunctionDef, 4, optional=True),
    _ApiObject("vector_jacobian_product", ast.FunctionDef, 4, optional=True),
    _ApiObject("abstract_eval", ast.FunctionDef, 1, optional=True),
)


def assert_relative_path(value: str) -> str:
    """Assert that a string encodes a relative path."""
    if Path(value).is_absolute():
        raise ValueError(f"value must be a relative path (got {value})")
    return value


RelativePath = Annotated[str, AfterValidator(assert_relative_path)]
StrictStr = Annotated[str, Strict()]


class TesseractBuildConfig(BaseModel):
    """Configuration options for building a Tesseract."""

    base_image: StrictStr = Field(
        "python:3.12-slim-bookworm",
        description="Base Docker image for the build. Must be Debian-based.",
    )
    target_platform: StrictStr = Field(
        "native",
        description=(
            "Target platform for the Docker build. Must be a valid Docker platform, "
            "or 'native' to build for the host platform. "
            "In general, images built for one platform will not run on another."
        ),
    )
    extra_packages: tuple[StrictStr, ...] = Field(
        (), description="Extra packages to install during build via apt-get."
    )
    package_data: tuple[tuple[RelativePath, StrictStr], ...] | None = Field(
        (),
        description=(
            "Additional files to copy into the Docker image, in the format ``(source, destination)``. "
            "Source paths are relative to the Tesseract directory."
        ),
    )
    custom_build_steps: tuple[StrictStr, ...] | None = Field(
        (),
        description=(
            "Custom steps to run during ``docker build`` (after everything else is installed). "
            "Example: ``[\"RUN echo 'Hello, world!'\"]``"
        ),
    )

    model_config = ConfigDict(extra="forbid")


# Allow None to be passed as a valid value for build_config, for example in YAML that comments out all options.
OptionalBuildConfig = Annotated[
    TesseractBuildConfig,
    BeforeValidator(lambda v: TesseractBuildConfig() if v is None else v),
]


class TesseractConfig(BaseModel):
    """Configuration options for Tesseracts. Defines valid options in ``tesseract_config.yaml``."""

    name: StrictStr = Field(..., description="Name of the Tesseract.")
    version: StrictStr = Field("0+unknown", description="Version of the Tesseract.")
    description: StrictStr = Field(
        "",
        description="Free-text description of what the Tesseract does.",
    )
    build_config: OptionalBuildConfig = Field(
        default_factory=TesseractBuildConfig,
        description="Configuration options for building the Tesseract.",
    )

    model_config = ConfigDict(extra="forbid")


class ValidationError(Exception):
    """Raised when inputs needed to build a tesseract are invalid."""

    pass


def _get_func_argnums(func: ast.FunctionDef) -> int:
    """Get the number of the arguments of a function node.

    See:
    https://docs.python.org/3/library/ast.html#ast.FunctionDef
    https://docs.python.org/3/library/ast.html#ast.arguments
    """
    func_args = func.args
    if func_args.kwonlyargs:
        raise ValidationError(
            f"Function {func.name} must not have keyword-only arguments"
        )
    if func_args.posonlyargs:
        raise ValidationError(
            f"Function {func.name} must not have positional-only arguments"
        )
    return len(func_args.args)


def validate_tesseract_api(src_dir: Path) -> None:
    """Check that given folder contains a Tesseract API that satisfies our constraints.

    This function does not return anything, but it raises `ValidationError` in
    case something goes wrong. In particular, we are checking that:
      *  The mandatory endpoints needed for a tesseract are actually
         implemented
      *  The implemented functions have the correct signature
      *  Both InputSchema and OutputSchema are `pydantic.BaseModel`s.

    Args:
        src_dir (Path): Path to the directory containing tesseract_api.py and tesseract_config.yaml.
    """
    tesseract_api_location = src_dir / "tesseract_api.py"
    config_location = src_dir / "tesseract_config.yaml"

    if not tesseract_api_location.exists():
        raise ValidationError(f"No file found at {tesseract_api_location}")

    if not config_location.exists():
        raise ValidationError(f"No file found at {config_location}")

    # Validate config
    try:
        get_config(src_dir)
    except PydanticValidationError as err:
        raise ValidationError(
            f"Invalid configuration in {config_location}: {err}"
        ) from err

    # Parse Tesseract API
    with open(tesseract_api_location) as f:
        tesseract_api_code = f.read()

    try:
        tesseract_api = ast.parse(tesseract_api_code)
    except SyntaxError as err:
        raise ValidationError(
            f"Syntax error in {tesseract_api_location}: {err}"
        ) from err

    # Check if expected attributes are defined
    toplevel_objects = {
        node.name: node for node in tesseract_api.body if hasattr(node, "name")
    }

    for obj in EXPECTED_OBJECTS:
        if obj.name not in toplevel_objects:
            if obj.optional:
                continue

            raise ValidationError(f"{obj.name} not defined in {tesseract_api_location}")

        if not isinstance(toplevel_objects[obj.name], obj.expected_type):
            raise ValidationError(
                f"{obj.name} is not a {obj.expected_type.__name__} in {tesseract_api_location}"
            )

        if obj.num_args is not None:
            if _get_func_argnums(toplevel_objects[obj.name]) != obj.num_args:
                raise ValidationError(f"{obj.name} must have {obj.num_args} arguments")

    # Check InputSchema and OutputSchema are pydantic BaseModels
    for schema in ("InputSchema", "OutputSchema"):
        obj = toplevel_objects[schema]
        if not obj.bases:
            subclass = None
        else:
            subclass = obj.bases[0].id

        if subclass != "BaseModel":
            raise ValidationError(
                f"{schema} must be a subclass of pydantic.BaseModel (got: {subclass})"
            )


def get_config(src_dir: Path) -> TesseractConfig:
    """Get configuration options from a tesseract_config.yaml file."""
    config_file = src_dir / "tesseract_config.yaml"

    if not config_file.exists():
        raise FileNotFoundError(f"No file found at {config_file}")

    with open(config_file) as f:
        config = yaml.safe_load(f)

    try:
        return TesseractConfig(**config)
    except PydanticValidationError as err:
        raise ValidationError(f"Invalid configuration: {err}") from err


def get_non_base_fields_in_tesseract_config() -> list[tuple[str, type]]:
    """Gets fields in Tesseract Config that are not a base fields."""
    base_fields = (str, int, float, bool, bytes)
    non_base_fields = []
    for field_name, field_info in TesseractConfig.model_fields.items():
        if field_info.annotation not in base_fields:
            non_base_fields.append((field_name, field_info.annotation))
    return non_base_fields
